//
//! @file CCtrlHexEdit.cpp
//! @copyright 1992 - 2011 Dennis Robinson (http://www.menasoft.com)
//
#include "StdAfx.h"
#include "CCtrlHexEdit.h"

#ifdef _MFC_VER

#include <ctype.h>
#include <afxdisp.h>
#include <afxole.h>

#include "../WinAPI/CWinHeap.h"
#include "../../GrayCore/CString.h"
#include "../../GrayCore/COSHandle.h"	// HANDLE_NULL

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

static const char s_hextable[16] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

#define TOHEX(a, b)	{*b++ = s_hextable[a >> 4];*b++ = s_hextable[a&0xf];}

//*******************************************
// CCtrlHexEdit

CCtrlHexEdit::CCtrlHexEdit()
	: m_topindex(0)
	, m_BytesPerRow(16)		// byte per row 
	, m_LinesPerPage(1)		// when i press page down.
	, m_bShowHex(true)
	, m_bShowAscii(true)
	, m_bShowAddress(true)
	, m_bAddressIsWide(true)		// 4/8 byte address
	, m_offAddress(0)
	, m_offHex(0)
	, m_offAscii(0)
	, m_bUpdate(true)			// update font info
	, m_bNoAddressChange(false)
	, m_currentMode(EDIT_NONE)
	, m_currentAddress(0)
	, m_bHalfPage(true)
	, m_selStart(0xffffffff)
	, m_selEnd(0xffffffff)
{
	m_editPos.x = m_editPos.y = 0;
	m_Font.CreateFont(-12, 0,0,0,0,0,0,0,0,0,0,0,0, _T("Courier New"));
}

CCtrlHexEdit::~CCtrlHexEdit()
{
}


BEGIN_MESSAGE_MAP(CCtrlHexEdit, CEdit)
	ON_WM_CONTEXTMENU()
	//{{AFX_MSG_MAP(CCtrlHexEdit)
	ON_WM_CHAR()
	ON_WM_KILLFOCUS()
	ON_WM_PAINT()
	ON_WM_SETFOCUS()
	ON_WM_SIZE()
	ON_WM_VSCROLL()
	ON_WM_HSCROLL()
	ON_WM_GETDLGCODE()
	ON_WM_ERASEBKGND()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONDBLCLK()
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONUP()
	ON_WM_KEYDOWN()
	ON_COMMAND(ID_EDIT_CLEAR, OnEditClear)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_COMMAND(ID_EDIT_CUT, OnEditCut)
	ON_COMMAND(ID_EDIT_PASTE, OnEditPaste)
	ON_COMMAND(ID_EDIT_SELECT_ALL, OnEditSelectAll)
	ON_COMMAND(ID_EDIT_UNDO, OnEditUndo)
	ON_WM_CREATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

//*******************************************
// CCtrlHexEdit message handlers


void CCtrlHexEdit::OnPaint() 
{
	CPaintDC pdc(this); // device context for painting

	CRect rc;
	GetClientRect(rc);
	
	CDC	dc;
	dc.CreateCompatibleDC(CDC::FromHandle(pdc.m_ps.hdc));
	CBitmap bm;

	bm.CreateCompatibleBitmap(CDC::FromHandle(pdc.m_ps.hdc), rc.Width(), rc.Height());
	dc.SelectObject(bm);

	CBrush b;
	b.CreateSolidBrush(RGB(0xff,0xff,0xff));
	dc.FillRect(rc, &b);

	ASSERT(m_currentAddress >= 0);
	ASSERT(m_topindex >= 0);

	dc.SelectObject(m_Font);
	int		height		= 0;

	dc.SetBoundsRect(&rc, DCB_DISABLE);

	BYTE* pData = GetDataPtr();
	if ( pData ) 
	{
		int x = rc.TopLeft().x;
		int y = rc.TopLeft().y;
		GChar_t buf[256];

		//
		// get char dimensions
		//
		if(m_bUpdate)
		{
			dc.GetCharWidth('0', '0', &m_nullWidth);
			CSize sz = dc.GetTextExtent(_T("0"), 1);
			m_lineHeight = sz.cy;
			
			m_offHex	= m_bShowAddress ? (m_bAddressIsWide ? m_nullWidth * 9 : m_nullWidth * 5) : 0;
			m_offAscii	= m_bShowAddress ? (m_bAddressIsWide ? m_nullWidth * 9 : m_nullWidth * 5) : 0;
			m_offAscii += m_bShowHex	 ? (m_BytesPerRow * 3 * m_nullWidth) : 0;

			m_LinesPerPage = rc.Height() / m_lineHeight;
			m_bHalfPage = false;
			if ( m_LinesPerPage * m_BytesPerRow > m_Data.get_DataLength())
			{
				m_LinesPerPage = (m_Data.get_DataLength() + (m_BytesPerRow/2)) / m_BytesPerRow ;
				if(m_Data.get_DataLength() % m_BytesPerRow != 0)
				{
					m_bHalfPage = true;
					m_LinesPerPage++;
				}
			}
			m_bUpdate = false;
			UpdateScrollBars();
		}

		TRACE( _T("%i %i\n"), m_topindex, m_selStart);
		
		height = rc.Height() / m_lineHeight;
		height *= m_lineHeight;

		if(m_bShowAddress)
		{
			GChar_t szFormat[8] = {'%','0','8','l','X','\0' };
			szFormat[2] = m_bAddressIsWide ? '8' : '4';
			int w = m_bAddressIsWide ? 8 : 4;
			y = 0;
			CRect rcd = rc;
			rcd.TopLeft().x = m_offAddress;
			for ( int i = m_topindex; 
				(i < m_Data.get_DataLength()) && (rcd.TopLeft().y < height); 
				i += m_BytesPerRow )
			{
				StrX::sprintfN( buf, _countof(buf), szFormat, i );
				dc.DrawText(buf, w, rcd, DT_LEFT|DT_TOP|DT_SINGLELINE|DT_NOPREFIX);
				rcd.TopLeft().y += m_lineHeight;
			}
		}
		if(m_bShowHex)
		{
			y = 0;
			CRect rcd = rc;
			rcd.TopLeft().x = x = m_offHex;

			if(m_selStart != 0xffffffff && (m_currentMode == EDIT_HIGH || m_currentMode == EDIT_LOW))
			{
				int	 i;
				int	 n = 0;
				int	 selStart = m_selStart, selEnd = m_selEnd;
				if(selStart > selEnd)
					selStart ^= selEnd ^= selStart ^= selEnd;

				for(i = m_topindex; (i < selStart) && (y < height); i++)
				{
					GChar_t* p = &buf[0];
					TOHEX(pData[i], p);
					*p++ = ' ';
					dc.TextOut(x, y, buf, 3);
					x += m_nullWidth * 3;
					n++;
					if(n == m_BytesPerRow)
					{
						n = 0;
						x = m_offHex;
						y += m_lineHeight;
					}
				}
				dc.SetTextColor(GetSysColor(COLOR_HIGHLIGHTTEXT));
				dc.SetBkColor(GetSysColor(COLOR_HIGHLIGHT));
				for(; (i < selEnd) && (i < m_Data.get_DataLength()) && (y < height); i++)
				{
					GChar_t* p = &buf[0];
					TOHEX(pData[i], p);
					*p++ = ' ';
					dc.TextOut(x, y, buf, 3);
					x += m_nullWidth * 3;
					n++;
					if(n == m_BytesPerRow)
					{
						n = 0;
						x = m_offHex;
						y += m_lineHeight;
					}
				}
				dc.SetTextColor(GetSysColor(COLOR_WINDOWTEXT));
				dc.SetBkColor(GetSysColor(COLOR_WINDOW));
				for(; (i < m_Data.get_DataLength()) && (y < height); i++)
				{
					GChar_t* p = &buf[0];
					TOHEX(pData[i], p);
					*p++ = ' ';
					dc.TextOut(x, y, buf, 3);
					x += m_nullWidth * 3;
					n++;
					if(n == m_BytesPerRow)
					{
						n = 0;
						x = m_offHex;
						y += m_lineHeight;
					}
				}
			}
			else
			{
				for(int	 i = m_topindex; (i < m_Data.get_DataLength()) && (rcd.TopLeft().y < height);)
				{
					GChar_t* p = &buf[0];
					int	 n = 0;
					for(; (n < m_BytesPerRow) && (i < m_Data.get_DataLength()); n++)
					{
						TOHEX(pData[i], p);
						*p++ = ' ';
						i++;
					}
					while(n < m_BytesPerRow)
					{
						*p++ = ' ';	*p++ = ' ';	*p++ = ' ';
						n++;
					}

					dc.DrawText(buf, m_BytesPerRow*3, rcd, DT_LEFT|DT_TOP|DT_SINGLELINE|DT_NOPREFIX);
					rcd.TopLeft().y += m_lineHeight;
				}
			}
		}
		if(m_bShowAscii)
		{
			y = 0;
			CRect rcd = rc;
			rcd.TopLeft().x = x = m_offAscii;
			if(m_selStart != 0xffffffff && m_currentMode == EDIT_ASCII)
			{
				int	 i;
				int	 n = 0;
				int	 selStart = m_selStart, selEnd = m_selEnd;
				if(selStart > selEnd)
					selStart ^= selEnd ^= selStart ^= selEnd;

				for(i = m_topindex; (i < selStart) && (y < height); i++)
				{
					buf[0] = StrChar::IsPrint(pData[i]) ? pData[i] : '.';
					dc.TextOut(x, y, buf, 1);
					x += m_nullWidth;
					n++;
					if(n == m_BytesPerRow)
					{
						n = 0;
						x = m_offAscii;
						y += m_lineHeight;
					}
				}
				dc.SetTextColor(GetSysColor(COLOR_HIGHLIGHTTEXT));
				dc.SetBkColor(GetSysColor(COLOR_HIGHLIGHT));
				for(; (i < selEnd) && (y < height); i++)
				{
					buf[0] = StrChar::IsPrint(pData[i]) ? pData[i] : '.';
					dc.TextOut(x, y, buf, 1);
					x += m_nullWidth;
					n++;
					if(n == m_BytesPerRow)
					{
						n = 0;
						x = m_offAscii;
						y += m_lineHeight;
					}
				}
				dc.SetTextColor(GetSysColor(COLOR_WINDOWTEXT));
				dc.SetBkColor(GetSysColor(COLOR_WINDOW));
				for(; (i < m_Data.get_DataLength()) && y < height; i++)
				{
					buf[0] = StrChar::IsPrint(pData[i]) ? pData[i] : '.';
					dc.TextOut(x, y, buf, 1);
					x += m_nullWidth;
					n++;
					if(n == m_BytesPerRow)
					{
						n = 0;
						x = m_offAscii;
						y += m_lineHeight;
					}
				}
			}
			else
			{
				for(int	 i = m_topindex; (i < m_Data.get_DataLength()) && (rcd.TopLeft().y < height);)
				{
					GChar_t* p = &buf[0];
					int	 n = 0;
					for(; (n < m_BytesPerRow) && (i < m_Data.get_DataLength()); n++)
					{
						*p++ = StrChar::IsPrint(pData[i]) ? pData[i] : '.';
						i++;
					}
					dc.DrawText(buf, n, rcd, DT_LEFT|DT_TOP|DT_SINGLELINE|DT_NOPREFIX);
					rcd.TopLeft().y += m_lineHeight;
				}
			}
		}
	}
	pdc.BitBlt(0, 0, rc.Width(), rc.Height(), &dc, 0, 0, SRCCOPY);
}

void CCtrlHexEdit::OnSetFocus(CWnd* pOldWnd) 
{
	if ( GetDataPtr() && !IsSelected())
	{
		if(m_editPos.x == 0 && m_bShowAddress)
			CreateAddressCaret();
		else
			CreateEditCaret();
		SetCaretPos(m_editPos);
		ShowCaret();
	}
	CWnd::OnSetFocus(pOldWnd);
}

void CCtrlHexEdit::OnKillFocus(CWnd* pNewWnd) 
{
	DestroyCaret();
	CWnd::OnKillFocus(pNewWnd);
}

void CCtrlHexEdit::OnSize(UINT nType, int cx, int cy) 
{
	CEdit::OnSize(nType, cx, cy);
}

void CCtrlHexEdit::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	// WARNING: nPos is really 16 bits !!!
	pScrollBar;
	if(!GetDataPtr())
		return;

	int oa = m_topindex;
	switch(nSBCode)
	{
		case SB_LINEDOWN:
			if(m_topindex < m_Data.get_DataLength() - m_LinesPerPage*m_BytesPerRow)
			{
				//ScrollWindow(0, -m_lineHeight, rc);
				m_topindex += m_BytesPerRow;
				RedrawWindow();
			}
			break;
		
		case SB_LINEUP:
			if(m_topindex > 0)
			{
				//ScrollWindow(0, m_lineHeight, rc);
				m_topindex -= m_BytesPerRow;
				RedrawWindow();
			}
			break;
		
		case SB_PAGEDOWN:
			if(m_topindex < m_Data.get_DataLength() - m_LinesPerPage*m_BytesPerRow)
			{
				m_topindex += m_BytesPerRow * m_LinesPerPage;
				if(m_topindex > m_Data.get_DataLength() - m_LinesPerPage*m_BytesPerRow)
					m_topindex = m_Data.get_DataLength() - m_LinesPerPage*m_BytesPerRow;
				RedrawWindow();
			}
			break;

		case SB_PAGEUP:
			if(m_topindex > 0)
			{
				m_topindex -= m_BytesPerRow * m_LinesPerPage;
				if(m_topindex < 0)
					m_topindex = 0;
				RedrawWindow();
			}
			break;

		case SB_THUMBTRACK:
			{
				SCROLLINFO si;
				si.cbSize = sizeof(SCROLLINFO);
				si.fMask = SIF_TRACKPOS;

				// Call GetScrollInfo to get current tracking 
				//    position in si.nTrackPos
				// Yep...Copied almost straight out of the Platform SDK =)

				if ( ! GetScrollInfo( SB_VERT, &si, SIF_TRACKPOS))
					return; 

				m_topindex = si.nTrackPos * m_BytesPerRow;
				RedrawWindow();
			}
			break;
	}
	::SetScrollPos(this->m_hWnd, SB_VERT, m_topindex / m_BytesPerRow, true);
	if(!m_bNoAddressChange && (m_topindex - oa) > 0) 
		m_currentAddress += (m_topindex - oa);
	RepositionCaret(m_currentAddress);
}

void CCtrlHexEdit::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	nSBCode;nPos; pScrollBar;
}

UINT CCtrlHexEdit::OnGetDlgCode() 
{
	return DLGC_WANTALLKEYS;
}

BOOL CCtrlHexEdit::PreCreateWindow(CREATESTRUCT& cs) 
{
	cs.style |= WS_HSCROLL|WS_VSCROLL;
	return CEdit::PreCreateWindow(cs);
}

BOOL CCtrlHexEdit::OnEraseBkgnd(CDC* pDC) 
{
	pDC;
	return true;
}

void CCtrlHexEdit::SetOptions(BOOL a, BOOL h, BOOL c, BOOL w)
{
	m_bShowHex		= h;
	m_bShowAscii	= c;
	m_bShowAddress	= a;
	m_bAddressIsWide= w;		// 4/8 byte address
	m_bUpdate		= true;
}

void CCtrlHexEdit::SetBPR(size_t bpr)
{
	m_BytesPerRow = bpr;
	m_bUpdate		= true;
}

void CCtrlHexEdit::OnLButtonDown(UINT nFlags, CPoint point) 
{
	SetFocus();
	if(GetDataPtr() == nullptr)
		return;

	if(nFlags & MK_SHIFT)
	{
		m_selStart = m_currentAddress;
	}
	CPointI pt = CalcPos(point.x, point.y);
	if(pt.x > -1)
	{
		m_editPos = pt;
		pt.x *= m_nullWidth;
		pt.y *= m_lineHeight;
		
		if(pt.x == 0 && m_bShowAddress)
			CreateAddressCaret();
		else
			CreateEditCaret();

		SetCaretPos(pt);
		if(nFlags & MK_SHIFT)
		{
			m_selEnd = m_currentAddress;
			if(m_currentMode == EDIT_HIGH || m_currentMode == EDIT_LOW)
				m_selEnd++;
			RedrawWindow();
		}
	}
	if(!(nFlags & MK_SHIFT))
	{
		if(::DragDetect( this->m_hWnd, point))
		{
			m_selStart = m_currentAddress;
			m_selEnd   = m_selStart;
			SetCapture();
		}
		else
		{
			BOOL bsel = m_selStart != 0xffffffff;

			m_selStart = 0xffffffff;
			m_selEnd   = 0xffffffff;
			if(bsel)
				RedrawWindow();
		}
	}
	if(!IsSelected())
	{
		ShowCaret();
	}
}

void CCtrlHexEdit::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	UNREFERENCED_PARAMETER(nFlags); UNREFERENCED_PARAMETER(point);
}

CPointI CCtrlHexEdit::CalcPos(int x, int y)
{
	y /= m_lineHeight;
	if(y < 0 || y > m_LinesPerPage)
		return CPointI(-1, -1);

	if(y * m_BytesPerRow > m_Data.get_DataLength())
		return CPointI(-1, -1);

	x += m_nullWidth;
	x /= m_nullWidth;

	int xp;

	if(m_bShowAddress && x <= (m_bAddressIsWide ? 8 : 4))
	{
		m_currentAddress = m_topindex + (m_BytesPerRow * y);
		m_currentMode = EDIT_NONE;
		return CPointI(0, y);
	}

	xp = (m_offHex  / m_nullWidth) + m_BytesPerRow * 3;
	if(m_bShowHex && x < xp)
	{
		if(x%3)
			x--;
		m_currentAddress = m_topindex + (m_BytesPerRow * y) + (x - (m_offHex  / m_nullWidth)) / 3;
		m_currentMode =  ((x%3) & 0x01) ? EDIT_LOW : EDIT_HIGH;
		return CPointI(x, y);
	}

	xp = (m_offAscii  / m_nullWidth) + m_BytesPerRow;
	if(m_bShowAscii && x <= xp)
	{
		m_currentAddress = m_topindex + (m_BytesPerRow * y) + (x - (m_offAscii  / m_nullWidth));
		m_currentMode = EDIT_ASCII;
		return CPointI(x, y);
	}

	return CPointI(-1,-1);
}

void CCtrlHexEdit::CreateAddressCaret()
{
	DestroyCaret();
	CreateSolidCaret(m_nullWidth * (m_bAddressIsWide ? 8 : 4), m_lineHeight);
}

void CCtrlHexEdit::CreateEditCaret()
{
	DestroyCaret();
	CreateSolidCaret(m_nullWidth, m_lineHeight);
}

void CCtrlHexEdit::OnMouseMove(UINT nFlags, CPoint point) 
{
	if(GetDataPtr()==nullptr)
		return;

	if(nFlags & MK_LBUTTON && m_selStart != 0xffffffff)
	{
		CRect rc;
		GetClientRect(&rc);
		if(!rc.PtInRect(point))
		{
			if(point.y < 0)
			{
				OnVScroll(SB_LINEUP, 0, nullptr);
				point.y = 0;
			}
			else if(point.y > rc.Height())
			{
				OnVScroll(SB_LINEDOWN, 0, nullptr);
				point.y = rc.Height() -1;
			}
		}

		//
		// we are selecting
		//
		int	 seo = m_selEnd;
		CPointI pt = CalcPos(point.x, point.y);
		if(pt.x > -1)
		{
			m_selEnd = m_currentAddress;
			if(m_currentMode == EDIT_HIGH || m_currentMode == EDIT_LOW)
				m_selEnd++;
		}
		if(IsSelected())
			DestroyCaret();

		if(seo != m_selEnd)
			RedrawWindow();
	}
}

void CCtrlHexEdit::UpdateScrollBars()
{
	SCROLLINFO si;

	si.cbSize = sizeof(SCROLLINFO);
	si.fMask = SIF_ALL;
	si.nMin = 0;
	si.nMax = (m_Data.get_DataLength() / m_BytesPerRow) - 1;
	si.nPage = m_LinesPerPage;
	si.nPos = m_topindex / m_BytesPerRow;

	::SetScrollInfo(this->m_hWnd, SB_VERT, &si, true);
	if(si.nMax > (int)si.nPage)
	{
		::EnableScrollBar(this->m_hWnd, SB_VERT, ESB_ENABLE_BOTH);
	}

	CRect rc;
	GetClientRect(&rc);

	si.nMin = 0;
	si.nMax =((m_bShowAddress ? (m_bAddressIsWide ? 8 : 4) : 0) +
			  (m_bShowHex ? m_BytesPerRow * 3 : 0) +
			  (m_bShowAscii ? m_BytesPerRow : 0)) * m_nullWidth;
	si.nPage = 1;
	si.nPos = 0;
	::SetScrollInfo(this->m_hWnd, SB_HORZ, &si, true);
}


inline BOOL CCtrlHexEdit::IsSelected()
{
	return m_selStart != 0xffffffff;
}

void CCtrlHexEdit::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if(IsSelected())
		ReleaseCapture();

	CWnd::OnLButtonUp(nFlags, point);
}

BYTE* CCtrlHexEdit::GetDataPtr() const
{
	return((BYTE*) m_Data.get_Data());
}

void CCtrlHexEdit::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	nFlags;nRepCnt;
	if(GetDataPtr()==nullptr)
		return;
	if(nChar == '\t')
		return;
	if(GetKeyState(VK_CONTROL) & 0x80000000)
	{
		switch(nChar)
		{
			case 0x03:
				if(IsSelected())
					OnEditCopy();
				return;
			case 0x16:
				OnEditPaste();
				return;
			case 0x18:
				if(IsSelected())
					OnEditCut();
				return;
			case 0x1a:
				OnEditUndo();
				return;
		}
	}

	if(nChar == 0x08)
	{
		if(m_currentAddress > 0)
		{
			m_currentAddress--;
			SelDelete(m_currentAddress, m_currentAddress+1);
			RepositionCaret(m_currentAddress);
			RedrawWindow();
		}
		return;
	}

	SetSel(-1, -1);
	switch(m_currentMode)
	{
		case EDIT_NONE:
			return;
		case EDIT_HIGH:
		case EDIT_LOW:
			if((nChar >= '0' && nChar <= '9') || (nChar >= 'a' && nChar <= 'f'))
			{
				UINT b = nChar - '0';
				if(b > 9) 
					b = 10 + nChar - 'a';

				if(m_currentMode == EDIT_HIGH)
				{
					GetDataPtr()[m_currentAddress] = (BYTE)((GetDataPtr()[m_currentAddress] & 0x0f) | (b << 4));
				}
				else
				{
					GetDataPtr()[m_currentAddress] = (BYTE)((GetDataPtr()[m_currentAddress] & 0xf0) | b);
				}
				Move(1,0);
			}
			break;
		case EDIT_ASCII:
			GetDataPtr()[m_currentAddress] = (BYTE)nChar;
			Move(1,0);
			break;
	}
	RedrawWindow();
}

void CCtrlHexEdit::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	nFlags;	nRepCnt;

	BOOL bShift = GetKeyState(VK_SHIFT) & 0x80000000;
	BOOL bac = m_bNoAddressChange;
	m_bNoAddressChange = true;
	switch(nChar)
	{
		case VK_DOWN:
			if(bShift)
			{
				if(!IsSelected())
				{
					m_selStart = m_currentAddress;
				}
				Move(0,1);
				m_selEnd   = m_currentAddress;
				if(m_currentMode == EDIT_HIGH || m_currentMode == EDIT_LOW)
					m_selEnd++;
				RedrawWindow();
				break;
			}
			else
				SetSel(-1, -1);
			Move(0,1);
			break;
		case VK_UP:
			if(bShift)
			{
				if(!IsSelected())
				{
					m_selStart = m_currentAddress;
				}
				Move(0,-1);
				m_selEnd   = m_currentAddress;
				RedrawWindow();
				break;
			}
			else
				SetSel(-1, -1);
			Move(0,-1);
			break;
		case VK_LEFT:
			if(bShift)
			{
				if(!IsSelected())
				{
					m_selStart = m_currentAddress;
				}
				Move(-1,0);
				m_selEnd   = m_currentAddress;
				RedrawWindow();
				break;
			}
			else
				SetSel(-1, -1);
			Move(-1,0);
			break;
		case VK_RIGHT:
			if(bShift)
			{
				if(!IsSelected())
				{
					m_selStart = m_currentAddress;
				}
				Move(1,0);
				m_selEnd   = m_currentAddress;
				if(m_currentMode == EDIT_HIGH || m_currentMode == EDIT_LOW)
					m_selEnd++;
				RedrawWindow();
				break;
			}
			else
				SetSel(-1, -1);
			Move(1,0);
			break;
		case VK_PRIOR:
			if(bShift)
			{
				if(!IsSelected())
				{
					m_selStart = m_currentAddress;
				}
				OnVScroll(SB_PAGEUP, 0, nullptr);
				Move(0,0);
				m_selEnd   = m_currentAddress;
				RedrawWindow();
				break;
			}
			else
				SetSel(-1, -1);
			OnVScroll(SB_PAGEUP, 0, nullptr);
			Move(0,0);
			break;
		case VK_NEXT:
			if(bShift)
			{
				if(!IsSelected())
				{
					m_selStart = m_currentAddress;
				}
				OnVScroll(SB_PAGEDOWN, 0, nullptr);
				Move(0,0);
				m_selEnd   = m_currentAddress;
				RedrawWindow();
				break;
			}
			else
				SetSel(-1, -1);
			OnVScroll(SB_PAGEDOWN, 0, nullptr);
			Move(0,0);
			break;
		case VK_HOME:
			if(bShift)
			{
				if(!IsSelected())
				{
					m_selStart = m_currentAddress;
				}
				if(GetKeyState(VK_CONTROL) & 0x80000000)
				{
					OnVScroll(SB_THUMBTRACK, 0, nullptr);
					Move(0,0);
				}
				else
				{
					m_currentAddress /= m_BytesPerRow;
					m_currentAddress *= m_BytesPerRow;
					Move(0,0);
				}
				m_selEnd   = m_currentAddress;
				RedrawWindow();
				break;
			}
			else
				SetSel(-1, -1);
			if(GetKeyState(VK_CONTROL) & 0x80000000)
			{
				OnVScroll(SB_THUMBTRACK, 0, nullptr);
				m_currentAddress = 0;
				Move(0,0);
			}
			else
			{
				m_currentAddress /= m_BytesPerRow;
				m_currentAddress *= m_BytesPerRow;
				Move(0,0);
			}
			break;
		case VK_END:
			if(bShift)
			{
				if(!IsSelected())
				{
					m_selStart = m_currentAddress;
				}
				if(GetKeyState(VK_CONTROL) & 0x80000000)
				{
					m_currentAddress = m_Data.get_DataLength()-1;
					OnVScroll(SB_THUMBTRACK, ((m_Data.get_DataLength()+(m_BytesPerRow/2)) / m_BytesPerRow) - m_LinesPerPage, nullptr);
					Move(0,0);
				}
				else
				{
					m_currentAddress /= m_BytesPerRow;
					m_currentAddress *= m_BytesPerRow;
					m_currentAddress += m_BytesPerRow - 1;
					if(m_currentAddress > m_Data.get_DataLength())
						m_currentAddress = m_Data.get_DataLength()-1;
					Move(0,0);
				}
				m_selEnd   = m_currentAddress;
				RedrawWindow();
				break;
			}
			else
				SetSel(-1, -1);
			if(GetKeyState(VK_CONTROL) & 0x80000000)
			{
				m_currentAddress = m_Data.get_DataLength()-1;
				if(m_bHalfPage)
					OnVScroll(SB_THUMBTRACK, 0, nullptr);
				else
					OnVScroll(SB_THUMBTRACK, ((m_Data.get_DataLength()+(m_BytesPerRow/2)) / m_BytesPerRow) - m_LinesPerPage, nullptr);
				Move(0,0);
			}
			else
			{
				m_currentAddress /= m_BytesPerRow;
				m_currentAddress *= m_BytesPerRow;
				m_currentAddress += m_BytesPerRow - 1;
				if(m_currentAddress > m_Data.get_DataLength())
					m_currentAddress = m_Data.get_DataLength()-1;
				Move(0,0);
			}
			break;
		case VK_INSERT:
			SelInsert(m_currentAddress, max(1, m_selEnd-m_selStart));
			RedrawWindow();
			break;
		case VK_DELETE:
			if(IsSelected())
			{
				OnEditClear();
			}
			else
			{
				SelDelete(m_currentAddress, m_currentAddress+1);
				RedrawWindow();
			}
			break;
		case '\t':
			switch(m_currentMode)
			{
				case EDIT_NONE:
					m_currentMode = EDIT_HIGH;
					break;
				case EDIT_HIGH:
				case EDIT_LOW:
					m_currentMode = EDIT_ASCII;
					break;
				case EDIT_ASCII:
					m_currentMode = EDIT_HIGH;
					break;
			}
			Move(0,0);
			break;

	}
	m_bNoAddressChange = bac;
}

void CCtrlHexEdit::Move(int x, int y)
{
	switch(m_currentMode)
	{
		case EDIT_NONE:
			return;
		case EDIT_HIGH:
			if(x != 0)
				m_currentMode = EDIT_LOW;
			if(x == -1)
				m_currentAddress --;
			m_currentAddress += y* m_BytesPerRow;
			break;
		case EDIT_LOW:
			if(x != 0)
				m_currentMode = EDIT_HIGH;
			if(x == 1)
				m_currentAddress++;
			m_currentAddress += y* m_BytesPerRow;
			break;
		case EDIT_ASCII:
			{
				m_currentAddress += x;
				m_currentAddress += y*m_BytesPerRow;
			}
			break;
	}
	if(m_currentAddress < 0)
		m_currentAddress = 0;

	if(m_currentAddress >= m_Data.get_DataLength())
	{
		m_currentAddress -= x;
		m_currentAddress -= y*m_BytesPerRow;
	}
	m_bNoAddressChange = true;
	if(m_currentAddress < m_topindex)
	{
		OnVScroll(SB_LINEUP, 0, nullptr);
	}
	if(m_currentAddress >= m_topindex + m_LinesPerPage*m_BytesPerRow)
	{
		OnVScroll(SB_LINEDOWN, 0, nullptr);
	}
	m_bNoAddressChange = false;
	//ScrollIntoView(m_currentAddress);
	RepositionCaret(m_currentAddress);
}

void CCtrlHexEdit::SetSel(int s, int e)
{
	DestroyCaret();
	m_selStart = s;
	m_selEnd = e;
	RedrawWindow();
	if(m_editPos.x == 0 && m_bShowAddress)
		CreateAddressCaret();
	else
		CreateEditCaret();
	SetCaretPos(m_editPos);
	ShowCaret();
}

void CCtrlHexEdit::RepositionCaret(int	 p)
{
	int x, y;

	y = (p - m_topindex) / m_BytesPerRow;
	x = (p - m_topindex) % m_BytesPerRow;

	switch(m_currentMode)
	{
		case EDIT_NONE:
			CreateAddressCaret();
			x = 0;
			break;
		case EDIT_HIGH:
			CreateEditCaret();
			x *= m_nullWidth*3;
			x += m_offHex;
			break;
		case EDIT_LOW:
			CreateEditCaret();
			x *= m_nullWidth*3;
			x += m_nullWidth;
			x += m_offHex;
			break;
		case EDIT_ASCII:
			CreateEditCaret();
			x *= m_nullWidth;
			x += m_offAscii;
			break;
	}
	m_editPos.x = x;
	m_editPos.y = y*m_lineHeight;
	CRect rc;
	GetClientRect(&rc);
	if(rc.PtInRect(m_editPos))
	{
		SetCaretPos(m_editPos);
		ShowCaret();
	}
}

void CCtrlHexEdit::ScrollIntoView(int p)
{
	if(p < m_topindex || p > m_topindex + m_LinesPerPage*m_BytesPerRow)
	{
		m_topindex = (p/m_BytesPerRow) * m_BytesPerRow;
		m_topindex -= (m_LinesPerPage / 3) * m_BytesPerRow;
		if(m_topindex < 0)
			m_topindex = 0;

		UpdateScrollBars();
		RedrawWindow();
	}
}

void CCtrlHexEdit::OnContextMenu(CWnd*, CPoint point)
{

	// CG: This block was added by the Pop-up Menu component
	{
		if (point.x == -1 && point.y == -1)
		{
			//keystroke invocation
			CRect rect;
			GetClientRect(rect);
			ClientToScreen(rect);

			point = rect.TopLeft();
			point += CPointI(5, 5);
		}

		CMenu menu;
		VERIFY(menu.LoadMenu(IDR_POPUP_HEX_EDIT));

		CMenu* pPopup = menu.GetSubMenu(0);
		ASSERT(pPopup != nullptr);

		pPopup->EnableMenuItem(ID_EDIT_UNDO, MF_GREYED|MF_DISABLED|MF_BYCOMMAND);
		if(!IsSelected())
		{
			pPopup->EnableMenuItem(ID_EDIT_CLEAR, MF_GREYED|MF_DISABLED|MF_BYCOMMAND);
			pPopup->EnableMenuItem(ID_EDIT_CUT, MF_GREYED|MF_DISABLED|MF_BYCOMMAND);
			pPopup->EnableMenuItem(ID_EDIT_COPY, MF_GREYED|MF_DISABLED|MF_BYCOMMAND);
		}
		{
			COleDataObject	obj;	
			if (obj.AttachClipboard()) 
			{
				if(!obj.IsDataAvailable(CF_TEXT) && 
					!obj.IsDataAvailable(RegisterClipboardFormat(_T("BinaryData"))))
					pPopup->EnableMenuItem(ID_EDIT_PASTE, MF_GREYED|MF_DISABLED|MF_BYCOMMAND);
			}
		}

		pPopup->TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y,
			this);
	}
}

void CCtrlHexEdit::OnEditClear() 
{
	m_currentAddress = m_selStart;
	SelDelete(m_selStart, m_selEnd);
	RepositionCaret(m_currentAddress);
	RedrawWindow();
}

void CCtrlHexEdit::OnEditCopy() 
{
	COleDataSource*		pSource = new COleDataSource();
	EmptyClipboard();
	BYTE* pData = GetDataPtr();
	ASSERT(pData!=nullptr);

	if(m_currentMode != EDIT_ASCII)
	{
		// CWinGlobalV
		int			dwLen = (m_selEnd-m_selStart);
		HGLOBAL		hMemb = ::GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE|GMEM_ZEROINIT, m_selEnd-m_selStart);
		HGLOBAL		hMema = ::GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE|GMEM_ZEROINIT, (dwLen) * 3);

		if (!hMema) 
			return;

		// copy binary
		LPBYTE	p = (BYTE*)::GlobalLock(hMemb);

		memcpy(p, pData+m_selStart, dwLen);
		::GlobalUnlock(hMemb);
		
		// copy ascii
		p = (BYTE*)::GlobalLock(hMema);
		for(int	 i = 0; i < dwLen;)
		{
			TOHEX(pData[m_selStart+i], p);
			*p++ = ' ';
			i++;
		}

		::GlobalUnlock(hMema);

		pSource->CacheGlobalData(RegisterClipboardFormat(_T("BinaryData")), hMemb);	
		pSource->CacheGlobalData(CF_TEXT, hMema);	
	}
	else
	{
		HGLOBAL				hMemb = ::GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE|GMEM_ZEROINIT, m_selEnd-m_selStart);
		HGLOBAL				hMema = ::GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE|GMEM_ZEROINIT, m_selEnd-m_selStart);
		if (!hMemb || !hMema) 
			return;

		// copy binary
		LPBYTE	p = (BYTE*)::GlobalLock(hMemb);
		int		dwLen = m_selEnd-m_selStart;

		memcpy(p, pData+m_selStart, dwLen);
		::GlobalUnlock(hMemb);
		
		// copy ascii
		p = (BYTE*)::GlobalLock(hMema);
		memcpy(p, pData+m_selStart, dwLen);
		for(int i = 0; i < dwLen; p++, i++)
			if(!StrChar::IsPrint(*p))
				*p = '.';
		::GlobalUnlock(hMema);

		pSource->CacheGlobalData(RegisterClipboardFormat( _T("BinaryData")), hMemb);	
		pSource->CacheGlobalData(CF_TEXT, hMema);	
	}
	pSource->SetClipboard();
}

void CCtrlHexEdit::OnEditCut() 
{
	OnEditCopy();
	SelDelete(m_selStart, m_selEnd);
	RedrawWindow();
}

void CCtrlHexEdit::OnEditPaste() 
{
	COleDataObject	obj;	
	if (obj.AttachClipboard()) 
	{
		HGLOBAL hmem = HANDLE_NULL;
		if (obj.IsDataAvailable(RegisterClipboardFormat( _T("BinaryData")))) 
		{
			hmem = obj.GetGlobalData(RegisterClipboardFormat( _T("BinaryData")));
		}	
		else if (obj.IsDataAvailable(CF_TEXT)) 
		{
			hmem = obj.GetGlobalData(CF_TEXT);
		}
		if(hmem!=HANDLE_NULL)
		{
			LPBYTE	p = (BYTE*)::GlobalLock(hmem);
			DWORD	dwLen = ::GlobalSize(hmem);
			int		insert;
			int		oa = m_currentAddress;
			
			NormalizeSel();
			if(m_selStart == 0xffffffff)
			{
				if(m_currentMode == EDIT_LOW)
					m_currentAddress++;
				insert = m_currentAddress;
				SelInsert(m_currentAddress, dwLen);
			}
			else
			{
				insert = m_selStart;
				SelDelete(m_selStart, m_selEnd);
				SelInsert(insert, dwLen);
				SetSel(-1, -1);
			}

			memcpy( GetDataPtr()+insert, p, dwLen);

			m_currentAddress = oa;
			RedrawWindow();
			::GlobalUnlock(hmem);
		}
	}
}

void CCtrlHexEdit::OnEditSelectAll() 
{
	m_selStart = 0;
	m_selEnd = m_Data.get_DataLength();
	DestroyCaret();
	RedrawWindow();
}

void CCtrlHexEdit::OnEditUndo() 
{
	// TODO: CCtrlHexEdit::OnEditUndo Add your command handler code here
	
}

void CCtrlHexEdit::NormalizeSel()
{
	if(m_selStart > m_selEnd)
		m_selStart ^= m_selEnd ^= m_selStart ^= m_selEnd;
}

void CCtrlHexEdit::SelDelete(int s, int e)
{
	ASSERT( s < m_Data.get_DataLength());

	CHeapBlockLen New;
	New.Alloc( m_Data.get_DataLength() - (e-s)+1 );

	BYTE* pNewData = (BYTE*) New.get_Data();
	memcpy( pNewData, GetDataPtr(), s );

	if(s < m_Data.get_DataLength()-(e-s)) 
	{
		memcpy(pNewData+s, GetDataPtr()+e, (m_Data.get_DataLength() -e));
	}

	m_Data.Free();
	m_Data = New;

	SetSel(-1, -1);
	if(m_currentAddress > m_Data.get_DataLength())
	{
		m_currentAddress = m_Data.get_DataLength();
		RepositionCaret(m_currentAddress);
	}
	m_bUpdate = true;
}

void CCtrlHexEdit::SelInsert(int s, int l)
{
	CHeapBlockLen New;
	New.Alloc( m_Data.get_DataLength() + l );
	BYTE* pNewData = (BYTE*) New.get_Data();

	memcpy(pNewData, GetDataPtr(), s);
	memcpy(pNewData+s+l, GetDataPtr()+s, (m_Data.get_DataLength()-s));

	m_Data.Free();
	m_Data = New;

	SetSel(-1, -1);
	m_bUpdate = true;
}

CSize CCtrlHexEdit::GetSel()
{
	return CSize(m_selStart, m_selEnd);
}

void CCtrlHexEdit::SetData(LPBYTE p, int len)
{
	m_Data.Alloc( len );
	memcpy( GetDataPtr(), p, len );

	SetSel(-1, -1);
	m_currentAddress = 0;
	m_editPos.x = m_editPos.y = 0;
	m_currentMode = EDIT_HIGH;
	m_topindex = 0;
	m_bUpdate = true;
}

int CCtrlHexEdit::GetData(LPBYTE p, int len)
{
	memcpy( p, GetDataPtr(), min(len, m_Data.get_DataLength()));
	return m_Data.get_DataLength();
}

int CCtrlHexEdit::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	lpCreateStruct->style |= WS_HSCROLL|WS_VSCROLL;

	if (CEdit::OnCreate(lpCreateStruct) == -1)
		return -1;

	SetFont(&m_Font);
	return 0;
}

#if 0

void CCtrlHexEdit::OnNcPaint() 
{
  // Display Caret at first data position 
  CreateEditCaret();
  m_currentAddress= 0;
  RepositionCaret(m_currentAddress);
  ShowCaret();
  
  CEdit::OnNcPaint(); // Be sure and call the parent class
}
#endif
#endif
