//
//! @file CWndPaneSlider.cpp
//! @copyright 1992 - 2011 Dennis Robinson (http://www.menasoft.com)
//
#include "StdAfx.h"
#include "CWndPaneSlider.h"

#ifdef _MFC_VER
#include "../../GrayCore/COSHandle.h"	// HANDLE_NULL
#include "../../GrayCore/CAppState.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//*******************************************
// CWndPaneSlider

CWndPaneSlider::CWndPaneSlider()
	: m_iPaneSlideX(-1)
	, m_pwndLeft(nullptr)
	, m_pwndRight(nullptr)
{
	m_iPaneSlideWidth = GetSystemMetrics( SM_CXSIZEFRAME );
}

CWndPaneSlider::~CWndPaneSlider()
{
}

void CWndPaneSlider::UpdatePositions( int cx, int cy )
{
	// update the positions given a resize of my parent window.
	// or move of the slider.
	if ( m_iPaneSlideX == -1 && cx > 1 )
	{
		m_iPaneSlideX = cx / 2;
	}

	int iSlideX;
	if ( m_pwndRight !=nullptr && m_pwndLeft !=nullptr ) 
		iSlideX = m_iPaneSlideX;
	else if ( m_pwndRight )
		iSlideX = 0;
	else 
		iSlideX = cx;

	if ( m_pwndLeft !=nullptr)
	{
		m_pwndLeft->MoveWindow( 0,0, iSlideX, cy );
	}
	
	MoveWindow( iSlideX, 0, m_iPaneSlideWidth, cy );

	if ( m_pwndRight !=nullptr)
	{
		m_pwndRight->MoveWindow( iSlideX+m_iPaneSlideWidth, 0, cx - iSlideX - m_iPaneSlideWidth, cy );
	}
}

void CWndPaneSlider::UpdatePositions()
{
	CRect rect;
	GetParent()->GetClientRect( &rect );
	UpdatePositions( rect.Width(), rect.Height());
}

void CWndPaneSlider::SetPaneWidth( CPoint pt )
{
	if ( m_pwndLeft == nullptr )
		return;
	if ( m_pwndRight == nullptr )
		return;

	CWnd* pParent = GetParent();
	ASSERT(pParent!=nullptr);
	pParent->ScreenToClient( &pt );
	CRect rect;
	pParent->GetClientRect( &rect );

	m_iPaneSlideX = pt.x;
	if ( m_iPaneSlideX < 0 ) 
		m_iPaneSlideX = 0;
	if ( m_iPaneSlideX > rect.Width()) 
		m_iPaneSlideX = rect.Width();

	// NOTE: Send a message that this has  changed
	GetParent()->SendMessage( WM_NOTIFY, m_NMHdr.idFrom, (LPARAM) &m_NMHdr ); // = (LPARAM) (LPNMHDR) pnmh; );  

	UpdatePositions( rect.Width(), rect.Height());
}

BEGIN_MESSAGE_MAP(CWndPaneSlider, CWnd)
	//{{AFX_MSG_MAP(CWndPaneSlider)
	ON_WM_NCHITTEST()
	ON_WM_NCLBUTTONDOWN()
	ON_WM_CREATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

//*******************************************
// CWndPaneSlider message handlers

#if _MSC_VER >= 1400
LRESULT CWndPaneSlider::OnNcHitTest(CPoint point) 
#else
UINT CWndPaneSlider::OnNcHitTest(CPoint point) 
#endif
{
	return HTLEFT;
	// return CWnd::OnNcHitTest(point);
}

void CWndPaneSlider::OnNcLButtonDown(UINT nHitTest, CPoint point) 
{
	// set capture to the window which received this message
	SetCapture();
	ASSERT(this == GetCapture());

	// get messages until capture lost or cancelled/accepted
	while ( GetCapture() == this )
	{
		MSG msg;
		if (!::GetMessage(&msg, WINHANDLE_NULL, 0, 0))
		{
			ReleaseCapture();
			AfxPostQuitMessage( (APP_EXITCODE_TYPE) msg.wParam);
			break;
		}
		switch (msg.message)
		{
		case WM_LBUTTONUP:
			// What are we on now ?
			ReleaseCapture();
			SetPaneWidth( msg.pt );
			break;
		case WM_MOUSEMOVE:
			// Change the mouse cursor depending on what we're on ?
			SetPaneWidth( msg.pt );
			break;
		case WM_RBUTTONDOWN:
		case WM_CANCELMODE:
			ReleaseCapture();
			break;
		case WM_KEYDOWN:
			if (msg.wParam == VK_ESCAPE)
			{
				ReleaseCapture();
				break;
			}
			break;
		default: // just dispatch rest of the messages
			::DispatchMessage(&msg);
			break;
		}
	}
	
	// CWnd::OnNcLButtonDown(nHitTest, point);
}

int CWndPaneSlider::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CWnd::OnCreate(lpCreateStruct) == -1)
		return -1;

    m_NMHdr.hwndFrom = this->m_hWnd;
    m_NMHdr.idFrom = GetDlgCtrlID();
	m_NMHdr.code = 0;
	return 0;
}
#endif
